wheel {
  background: color-mix(in srgb, currentColor 8%, transparent);
  border-radius: 100%;
  margin: 40px;
  min-width: 280px;
  min-height: 280px;
}

wheel > button {
  background-color: var(--window-bg-color);
  border: 1px solid color-mix(in srgb, currentColor 15%, transparent);
  padding: 3px;
}

wheel > button.playing {
  background-color: var(--accent-bg-color);
  color: var(--accent-fg-color);
}
